# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Liblarch - a library to handle directed acyclic graphs
# Copyright (c) 2011-2012 - Lionel Dricot & Izidor Matušov
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import threading

from gi.repository import GObject


class SyncQueue(object):
    """ Synchronized queue for processing requests"""

    def __init__(self):
        """ Initialize synchronized queue.

        @param callback - function for processing requests"""
        self._low_queue = []
        self._queue = []
        self._vip_queue = []
        self._handler = None
        self._lock = threading.Lock()
        self._origin_thread = threading.current_thread()

        self.count = 0

    def process_queue(self):
        """ Process requests from queue """
        for action in self.process():
            func = action[0]
            func(*action[1:])
        # return True to process other requests as well
        return True

    def push(self, *element, **kwargs):
        """ Add a new element to the queue.

        Process actions from the same thread as the thread which created
        this queue immediately. What does it mean? When I use liblarch
        without threads, all actions are processed immediately. In GTG,
        this queue is created by the main thread which process GUI. When
        GUI callback is triggered, process those actions immediately because
        no protection is needed. However, requests from synchronization
        services are put in the queue.

        Application can choose which kind of priority should have an update.
        If the request is not in the queue of selected priority, add it and
        setup callback.
        """

        if self._origin_thread == threading.current_thread():
            func = element[0]
            func(*element[1:])
            return

        priority = kwargs.get('priority')
        if priority == 'low':
            queue = self._low_queue
        elif priority == 'high':
            queue = self._vip_queue
        else:
            queue = self._queue

        self._lock.acquire()
        if element not in queue:
            queue.append(element)
            if self._handler is None:
                self._handler = GObject.idle_add(self.process_queue)

        self._lock.release()

    def process(self):
        """ Return elements to process

        At the moment, it returns just one element. In the future more
        elements may be better to return (to speed it up).

        If there is no request left, disable processing. """

        self._lock.acquire()
        if len(self._vip_queue) > 0:
            toreturn = [self._vip_queue.pop(0)]
        elif len(self._queue) > 0:
            toreturn = [self._queue.pop(0)]
        elif len(self._low_queue) > 0:
            toreturn = [self._low_queue.pop(0)]
        else:
            toreturn = []

        if (len(self._queue) == 0 and
                len(self._vip_queue) == 0 and
                len(self._low_queue) == 0 and
                self._handler is not None):
            GObject.source_remove(self._handler)
            self._handler = None
        self._lock.release()
        return toreturn



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Liblarch - a library to handle directed acyclic graphs
# Copyright (c) 2011-2012 - Lionel Dricot & Izidor Matušov
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# ---------------------------------------------------------------------------


class ViewCount(object):
    def __init__(self, tree, fbank, name=None, refresh=True):
        self.initialized = False
        self.ncount = {}
        self.tree = tree
        self.tree.register_callback("node-added", self.__modify)
        self.tree.register_callback("node-modified", self.__modify)
        self.tree.register_callback("node-deleted", self.__delete)

        self.fbank = fbank
        self.name = name

        self.applied_filters = []
        self.nodes = []
        self.cllbcks = []

        if refresh:
            self.__refresh()

    def __refresh(self):
        for node in self.tree.get_all_nodes():
            self.__modify(node)
        self.initialized = True

    def apply_filter(self, filter_name, refresh=True):
        if self.fbank.has_filter(filter_name):
            if filter_name not in self.applied_filters:
                self.applied_filters.append(filter_name)
                if refresh:
                    # If we are not initialized, we need to refresh with
                    # all existing nodes
                    if self.initialized:
                        for n in list(self.nodes):
                            self.__modify(n)
                    else:
                        self.__refresh()
        else:
            print("There's no filter called %s" % filter_name)

    def unapply_filter(self, filter_name):
        if filter_name in self.applied_filters:
            self.applied_filters.remove(filter_name)
            for node in self.tree.get_all_nodes():
                self.__modify(node)

    # there's only one callback: "modified"
    def register_cllbck(self, func):
        if func not in self.cllbcks:
            self.cllbcks.append(func)

    def unregister_cllbck(self, func):
        if func in self.cllbacks:
            self.cllbacks.remove(func)

    def get_n_nodes(self):
        return len(self.nodes)

    def modify(self, nid):
        """ Allow external update of a given node """
        self.__modify(nid)

    def __modify(self, nid):
        displayed = True
        for filtname in self.applied_filters:
            filt = self.fbank.get_filter(filtname)
            displayed &= filt.is_displayed(nid)
        if displayed:
            self.__add(nid)
        else:
            self.__delete(nid)

    def __delete(self, nid):
        if nid in self.nodes:
            self.nodes.remove(nid)
            self.__callback()

    def __add(self, nid):
        if nid not in self.nodes:
            self.nodes.append(nid)
            self.__callback()

    def __callback(self):
        for c in self.cllbcks:
            c()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Liblarch - a library to handle directed acyclic graphs
# Copyright (c) 2011-2012 - Lionel Dricot & Izidor Matušov
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

from .filters_bank import FiltersBank
from .tree import MainTree
from .treenode import _Node
from .viewcount import ViewCount
from .viewtree import ViewTree

# API version of liblarch.
# Your application is compatible if the major version number match liblarch's
# one and if your minor version number is inferior to liblarch's one.
#
# The major number is incremented if an existing method is removed or modified
# The minor number is incremented when a method is added to the API
API = "3.0"


def is_compatible(request):
    major, minor = [int(i) for i in request.split(".")]
    current_ma, current_mi = [int(i) for i in API.split(".")]
    return major == current_ma and minor <= current_mi


class TreeNode(_Node):
    """ The public interface for TreeNode
    """
    def __init__(self, node_id, parent=None):
        _Node.__init__(self, node_id, parent)

    def _set_tree(tree):
        print("_set_tree is not part of the API")


class Tree(object):
    """ A thin wrapper to MainTree that adds filtering capabilities.
    It also provides a few methods to operate complex operation on the
    MainTree (e.g, move_node) """

    def __init__(self):
        """ Creates MainTree which wraps and a main view without filters """
        self.__tree = MainTree()
        self.__fbank = FiltersBank(self.__tree)
        self.__views = {}
        self.__viewscount = {}
        self.__views['main'] = ViewTree(
            self, self.__tree, self.__fbank, static=True)

    # HANDLE NODES ############################################################
    def get_node(self, node_id):
        """ Returns the object of node.
        If the node does not exists, a ValueError is raised. """
        return self.__tree.get_node(node_id)

    def has_node(self, node_id):
        """ Does the node exists in this tree? """
        return self.__tree.has_node(node_id)

    def add_node(self, node, parent_id=None, priority="low"):
        """ Add a node to tree. If parent_id is set, put the node as a child of
        this node, otherwise put it as a child of the root node."""
        self.__tree.add_node(node, parent_id, priority)

    def del_node(self, node_id, recursive=False):
        """ Remove node from tree and return whether it was successful or not
        """
        return self.__tree.remove_node(node_id, recursive)

    def refresh_node(self, node_id, priority="low"):
        """ Send a request for updating the node """
        self.__tree.modify_node(node_id, priority)

    def refresh_all(self):
        """ Refresh all nodes """
        self.__tree.refresh_all()

    def move_node(self, node_id, new_parent_id=None):
        """ Move the node to a new parent (dismissing all other parents)
        use pid None to move it to the root """
        if self.has_node(node_id):
            node = self.get_node(node_id)
            node.set_parent(new_parent_id)
            toreturn = True
        else:
            toreturn = False

        return toreturn

    def add_parent(self, node_id, new_parent_id=None):
        """ Add the node to a new parent. Return whether operation was
        successful or not. If the node does not exists, return False """

        if self.has_node(node_id):
            node = self.get_node(node_id)
            return node.add_parent(new_parent_id)
        else:
            return False

    # VIEWS ###################################################################
    def get_main_view(self):
        """ Return the special view "main" which is without any filters on it.
        """
        return self.__views['main']

    def get_viewtree(self, name=None, refresh=True):
        """ Returns a viewtree by the name:
          * a viewtree with that name exists => return it
          * a viewtree with that name does not exist => create a new one and
            return it
          * name is None => create an anonymous tree (do not remember it)

        If refresh is False, the view is not initialized. This is useful as
        an optimization if you plan to apply a filter.
        """

        if name is not None and name in self.__views:
            view_tree = self.__views[name]
        else:
            view_tree = ViewTree(
                self, self.__tree, self.__fbank, name=name, refresh=refresh)
            if name is not None:
                self.__views[name] = view_tree
        return view_tree

    def get_viewcount(self, name=None, refresh=True):
        if name is not None and name in self.__viewscount:
            view_count = self.__viewscount[name]
        else:
            view_count = ViewCount(
                self.__tree, self.__fbank, name=name, refresh=refresh)
            if name is not None:
                self.__viewscount[name] = view_count
        return view_count

    # FILTERS #################################################################
    def list_filters(self):
        """ Return a list of all available filters by name """
        return self.__fbank.list_filters()

    def add_filter(self, filter_name, filter_func, parameters=None):
        """ Adds a filter to the filter bank.

        @filter_name : name to give to the filter
        @filter_func : the function that will filter the nodes
        @parameters : some default parameters fot that filter
        Return True if the filter was added
        Return False if the filter_name was already in the bank
        """
        return self.__fbank.add_filter(filter_name, filter_func, parameters)

    def remove_filter(self, filter_name):
        """ Remove a filter from the bank. Only custom filters that were
        added here can be removed. Return False if the filter was not removed.
        """
        return self.__fbank.remove_filter(filter_name)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Liblarch - a library to handle directed acyclic graphs
# Copyright (c) 2011-2012 - Lionel Dricot & Izidor Matušov
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import gi  # noqa
gi.require_version("Gtk", "3.0")  # noqa
from gi.repository import Gtk, Gdk
from gi.repository import GObject

from .treemodel import TreeModel


# Useful for debugging purpose.
# Disabling that will disable the TreeModelSort on top of our TreeModel
ENABLE_SORTING = True
USE_TREEMODELFILTER = False


BRITGHTNESS_CACHE = {}


def brightness(color_str):
    """ Compute brightness of a color on scale 0-1

    Courtesy to
    http://stackoverflow.com/questions/596216/formula-to-determine-brightness-of-rgb-color
    """
    global BRITGHTNESS_CACHE

    if color_str not in BRITGHTNESS_CACHE:
        c = Gdk.color_parse(color_str)
        brightness = (0.2126 * c.red + 0.7152 * c.green + 0.0722 * c.blue) / 65535.0
        BRITGHTNESS_CACHE[color_str] = brightness
    return BRITGHTNESS_CACHE[color_str]


class TreeView(Gtk.TreeView):
    """ Widget which display LibLarch FilteredTree.

    This widget extends Gtk.TreeView by several features:
      * Drag'n'Drop support
      * Sorting support
      * separator rows
      * background color of a row
      * selection of multiple rows
    """

    __string_signal__ = (GObject.SignalFlags.RUN_FIRST, None, (str, ))
    __gsignals__ = {
        'node-expanded': __string_signal__,
        'node-collapsed': __string_signal__,
    }

    def __init__(self, tree, description):
        """ Build the widget

        @param  tree - LibLarch FilteredTree
        @param  description - definition of columns.

        Parameters of description dictionary for a column:
          * value => (type of values, function for generating value from
                      a node)
          * renderer => (renderer_attribute, renderer object)

          Optional:
          * order => specify order of column otherwise use natural order
          * expandable => is the column expandable?
          * resizable => is the column resizable?
          * visible => is the column visible?
          * title => title of column
          * new_colum => do not create a separate column, just continue with
                         the previous one (this can be used to create
                         columns without borders)
          * sorting => allow default sorting on this column
          * sorting_func => use special function for sorting on this func

        Example of columns descriptions:
        description = { 'title': {
                'value': [str, self.task_title_column],
                'renderer': ['markup', Gtk.CellRendererText()],
                'order': 0
            }}
        """
        GObject.GObject.__init__(self)
        self.columns = {}
        self.sort_col = None
        self.sort_order = Gtk.SortType.ASCENDING
        self.bg_color_column = None
        self.separator_func = None

        self.dnd_internal_target = ''
        self.dnd_external_targets = {}

        # Sort columns
        self.order_of_column = []
        last = 9999
        for col_name in description:
            desc = description[col_name]
            order = desc.get('order', last)
            last += 1
            self.order_of_column.append((order, col_name))

        types = []
        sorting_func = []
        # Build the first coulumn if user starts with new_colum=False
        col = Gtk.TreeViewColumn()

        # Build columns according to the order
        for col_num, (order_num, col_name) in enumerate(
                sorted(self.order_of_column), 1):
            desc = description[col_name]
            types.append(desc['value'])

            expand = desc.get('expandable', False)
            resizable = desc.get('resizable', True)
            visible = desc.get('visible', True)

            if 'renderer' in desc:
                rend_attribute, renderer = desc['renderer']
            else:
                rend_attribute = 'markup'
                renderer = Gtk.CellRendererText()

            # If new_colum=False, do not create new column, use the previous
            # one. It will create columns without borders.
            if desc.get('new_column', True):
                col = Gtk.TreeViewColumn()
                newcol = True
            else:
                newcol = False
            col.set_visible(visible)

            if 'title' in desc:
                col.set_title(desc['title'])

            col.pack_start(renderer, expand=expand)
            col.add_attribute(renderer, rend_attribute, col_num)
            col.set_resizable(resizable)
            col.set_expand(expand)

            # Allow to set background color
            col.set_cell_data_func(renderer, self._celldatafunction)

            if newcol:
                self.append_column(col)
            self.columns[col_name] = (col_num, col)

            if ENABLE_SORTING:
                if 'sorting' in desc:
                    # Just allow sorting and use default comparing
                    self.sort_col = desc['sorting']
                    sort_num, sort_col = self.columns[self.sort_col]
                    col.set_sort_column_id(sort_num)

                if 'sorting_func' in desc:
                    # Use special funcion for comparing, e.g. dates
                    sorting_func.append((col_num, col, desc['sorting_func']))

        self.basetree = tree
        # Build the model around LibLarch tree
        self.basetreemodel = TreeModel(tree, types)
        # Applying an intermediate treemodelfilter, for debugging purpose
        if USE_TREEMODELFILTER:
            treemodelfilter = self.basetreemodel.filter_new()
        else:
            treemodelfilter = self.basetreemodel

        # Apply TreeModelSort to be able to sort
        if ENABLE_SORTING:
            self.treemodel = self.basetreemodel
            for col_num, col, sort_func in sorting_func:
                self.treemodel.set_sort_func(
                    col_num, self._sort_func, sort_func)
                col.set_sort_column_id(col_num)
        else:
            self.treemodel = treemodelfilter

        self.set_model(self.treemodel)

        self.expand_all()
        self.show()

        self.collapsed_paths = []
        self.connect('row-expanded', self.__emit, 'expanded')
        self.connect('row-collapsed', self.__emit, 'collapsed')
        self.treemodel.connect('row-has-child-toggled', self.on_child_toggled)

    def __emit(self, sender, iter, path, data):
        """ Emit expanded/collapsed signal """
        # recreating the path of the collapsed node
        ll_path = ()
        i = 1
        path = path.get_indices()
        while i <= len(path):
            temp_path = Gtk.TreePath(":".join(str(n) for n in path[:i]))
            temp_iter = self.treemodel.get_iter(temp_path)
            ll_path += (self.treemodel.get_value(temp_iter, 0), )
            i += 1
        if data == 'expanded':
            self.emit('node-expanded', ll_path)
        elif data == 'collapsed':
            self.emit('node-collapsed', ll_path)

    def on_child_toggled(self, treemodel, path, iter, param=None):
        """ Expand row """
        # is the toggled node in the collapsed paths?
        collapsed = False
        nid = treemodel.get_value(iter, 0)
        while iter and not collapsed:
            for c in self.collapsed_paths:
                if c[-1] == nid:
                    collapsed = True
            iter = treemodel.iter_parent(iter)
        if not self.row_expanded(path) and not collapsed:
            self.expand_row(path, True)

    def expand_node(self, llpath):
        """ Expand the children of a node. This is not recursive """
        self.collapse_node(llpath, collapsing_method=self.expand_one_row)

    def expand_one_row(self, p):
        # We have to set the "open all" parameters
        self.expand_row(p, False)

    def collapse_node(self, llpath, collapsing_method=None):
        """ Hide children of a node

        This method is needed for "rember collapsed nodes" feature of GTG.
        Transform node_id into paths and those paths collapse. By default all
        children are expanded (see self.expand_all())

        @parameter llpath - LibLarch path to the node. Node_id is extracted
            as the last parameter and then all instances of that node are
            collapsed. For retro-compatibility, we take llpath instead of
            node_id directly"""
        if not collapsing_method:
            collapsing_method = self.collapse_row
        node_id = llpath[-1].strip("'")
        if not node_id:
            raise Exception('Missing node_id in path %s' % str(llpath))

        schedule_next = True
        for path in self.basetree.get_paths_for_node(node_id):
            iter = self.basetreemodel.my_get_iter(path)
            if iter is None:
                continue

            target_path = self.basetreemodel.get_path(iter)
            if self.basetreemodel.get_value(iter, 0) == node_id:
                collapsing_method(target_path)
                self.collapsed_paths.append(path)
                schedule_next = False

        if schedule_next:
            self.basetree.queue_action(
                node_id, self._collapse_node_retry,
                param=(llpath, collapsing_method))

    def _collapse_node_retry(self, param):
        """
        Node to be collapsed/expand found, so re-try now with correct
        and preserved parameters.
        """
        self.collapse_node(param[0], collapsing_method=param[1])

    def show(self):
        """ Shows the TreeView and connect basetreemodel to LibLarch """
        Gtk.TreeView.show(self)
        self.basetreemodel.connect_model()

    def get_columns(self):
        """ Return the list of columns name """
        return list(self.columns.keys())

    def set_main_search_column(self, col_name):
        """ Set search column for GTK integrate search
        This is just wrapper to use internal representation of columns"""
        col_num, col = self.columns[col_name]
        self.set_search_column(col_num)

    def set_expander_column(self, col_name):
        """ Set expander column (that which expands through free space)
        This is just wrapper to use internal representation of columns"""
        col_num, col = self.columns[col_name]
        self.set_property("expander-column", col)

    def set_sort_column(self, col_name, order=Gtk.SortType.ASCENDING):
        """ Select column to sort by it by default """
        if ENABLE_SORTING:
            self.sort_col = col_name
            self.sort_order = order
            col_num, col = self.columns[col_name]
            self.treemodel.set_sort_column_id(col_num, order)

    def get_sort_column(self):
        """ Get sort column """
        if ENABLE_SORTING:
            return self.sort_col, self.sort_order

    def set_col_visible(self, col_name, visible):
        """ Set visiblity of column.
        Allow to hide/show certain column """
        col_num, col = self.columns[col_name]
        col.set_visible(visible)

    def set_col_resizable(self, col_name, resizable):
        """ Allow/forbid column to be resizable """
        col_num, col = self.columns[col_name]
        col.set_resizable(resizable)

    def set_bg_color(self, color_func, color_column):
        """ Set which column and function for generating background color

        Function should be in format func(node, default_color)
        """

        def closure_default_color(func, column):
            """ Set default color to the function.

            Transform function from func(node, default_color) into func(node).
            Default color is computed based on some GTK style magic. """
            style = column.get_tree_view().get_style_context()
            color = style.get_background_color(Gtk.StateFlags.NORMAL)
            default = color.to_color()
            return lambda node: func(node, default)

        if color_column in self.columns:
            self.bg_color_column, column = self.columns[color_column]
            func = closure_default_color(color_func, column)
            self.treemodel.set_column_function(self.bg_color_column, func)
        else:
            raise ValueError(
                "There is no colum %s to use to set color" % color_column)

    def _sort_func(self, model, iter1, iter2, func=None):
        """ Sort two iterators by function which gets node objects.
        This is a simple wrapper which prepares node objects and then
        call comparing function. In other case return default value -1
        """
        node_id_a = model.get_value(iter1, 0)
        node_id_b = model.get_value(iter2, 0)
        if node_id_a and node_id_b and func:
            id, order = self.treemodel.get_sort_column_id()
            node_a = self.basetree.get_node(node_id_a)
            node_b = self.basetree.get_node(node_id_b)
            sort = func(node_a, node_b, order)
        else:
            sort = -1
        return sort

    def _celldatafunction(self, column, cell, model, myiter, user_data):
        """ Determine background color for cell

        Requirements: self.bg_color_column must be set
        (see self.set_bg_color())

        Set background color based on a certain column value.
        """
        if self.bg_color_column is None:
            return

        if myiter and model.iter_is_valid(myiter):
            color = model.get_value(myiter, self.bg_color_column)
        else:
            color = None

        if isinstance(cell, Gtk.CellRendererText):
            if color is not None and brightness(color) < 0.5:
                cell.set_property("foreground", '#FFFFFF')
            else:
                # Otherwise unset foreground color
                cell.set_property("foreground-set", False)

        cell.set_property("cell-background", color)

    # DRAG-N-DROP functions #####################################

    def set_dnd_name(self, dndname):
        """ Sets Drag'n'Drop name and initialize Drag'n'Drop support

        If ENABLE_SORTING, drag_drop signal must be handled by this widget."""
        self.dnd_internal_target = dndname
        self.__init_dnd()
        self.connect('drag_data_get', self.on_drag_data_get)
        self.connect('drag_data_received', self.on_drag_data_received)

    def set_dnd_external(self, sourcename, func):
        """ Add a new external target and initialize Drag'n'Drop support"""
        i = 1
        while i in self.dnd_external_targets:
            i += 1
        self.dnd_external_targets[i] = [sourcename, func]
        self.__init_dnd()

    def __init_dnd(self):
        """ Initialize Drag'n'Drop support

        Firstly build list of DND targets:
            * name
            * scope - just the same widget / same application
            * id

        Enable DND by calling enable_model_drag_dest(),
        enable_model-drag_source()

        It didnt use support from Gtk.Widget(drag_source_set(),
        drag_dest_set()). To know difference, look in PyGTK FAQ:
        http://faq.pyGtk.org/index.py?file=faq13.033.htp&req=show
        """
        self.defer_select = False

        if self.dnd_internal_target == '':
            error = 'Cannot initialize DND without a valid name\n'
            error += 'Use set_dnd_name() first'
            raise Exception(error)

        dnd_targets = [(
            self.dnd_internal_target, Gtk.TargetFlags.SAME_WIDGET, 0)]
        for target in self.dnd_external_targets:
            name = self.dnd_external_targets[target][0]
            dnd_targets.append((name, Gtk.TargetFlags.SAME_APP, target))

        self.enable_model_drag_source(
            Gdk.ModifierType.BUTTON1_MASK,
            dnd_targets,
            Gdk.DragAction.DEFAULT | Gdk.DragAction.MOVE)

        self.enable_model_drag_dest(
            dnd_targets, Gdk.DragAction.DEFAULT | Gdk.DragAction.MOVE)

    def on_drag_data_get(self, treeview, context, selection, info, timestamp):
        """ Extract data from the source of the DnD operation.

        Serialize iterators of selected tasks in format
        <iter>, <iter>, ..., <iter> and set it as parameter of DND """

        treeselection = treeview.get_selection()
        model, paths = treeselection.get_selected_rows()
        iters = [model.get_iter(path) for path in paths]
        iter_str = ','.join(model.get_string_from_iter(iter) for iter in iters)
        selection.set(selection.get_target(), 0, iter_str.encode('ascii'))

    def on_drag_data_received(self, treeview, context, x, y, selection, info,
                              timestamp):
        """ Handle a drop situation.

        First of all, we need to get id of node which should accept
        all draged nodes as their new children. If there is no node,
        drop to root node.

        Deserialize iterators of dragged nodes (see self.on_drag_data_get())
        Info parameter determines which target was used:
            * info == 0 => internal DND within this TreeView
            * info > 0 => external DND

        In case of internal DND we just use Tree.move_node().
        In case of external DND we call function associated with that DND
        set by self.set_dnd_external()
        """
        # TODO: it should be configurable for each TreeView if you want:
        # 0 : no drag-n-drop at all
        # 1 : drag-n-drop move the node
        # 2 : drag-n-drop copy the node

        model = treeview.get_model()
        drop_info = treeview.get_dest_row_at_pos(x, y)
        if drop_info:
            path, position = drop_info
            iter = model.get_iter(path)
            # Must add the task to the parent of the task situated
            # before/after
            if position == Gtk.TreeViewDropPosition.BEFORE or\
               position == Gtk.TreeViewDropPosition.AFTER:
                # Get sibling parent
                destination_iter = model.iter_parent(iter)
            else:
                # Must add task as a child of the dropped-on iter
                # Get parent
                destination_iter = iter

            if destination_iter:
                destination_tid = model.get_value(destination_iter, 0)
            else:
                # it means we have drag-n-dropped above the first task
                # we should consider the destination as a root then.
                destination_tid = None
        else:
            # Must add the task to the root
            # Parent = root => iter=None
            destination_tid = None

        tree = self.basetree.get_basetree()

        # Get dragged iter as a TaskTreeModel iter
        # If there is no selected task (empty selection.data),
        # explictly skip handling it (set to empty list)
        data = selection.get_data()
        if data == '':
            iters = []
        else:
            iters = data.decode().split(',')

        dragged_iters = []
        for iter in iters:
            if info == 0:
                try:
                    dragged_iters.append(model.get_iter_from_string(iter))
                except ValueError:
                    # I hate to silently fail but we have no choice.
                    # It means that the iter is not good.
                    # Thanks shitty gtk API for not allowing us to test
                    # the string
                    dragged_iter = None

            # Handle drag from one widget to another (ex: treeview to treeview)
            elif info in self.dnd_external_targets and destination_tid:
                f = self.dnd_external_targets[info][1]
                src_model = Gtk.drag_get_source_widget(context).get_model()
                dragged_iters.append(src_model.get_iter_from_string(iter))

        for dragged_iter in dragged_iters:
            if info == 0:
                if dragged_iter and model.iter_is_valid(dragged_iter):
                    dragged_tid = model.get_value(dragged_iter, 0)
                    try:
                        tree.move_node(
                            dragged_tid, new_parent_id=destination_tid)
                    except Exception as e:
                        print('Problem with dragging: %s' % e)

            # Handle inter-widget Drag'n'Drop again (like in the previous loop)
            elif info in self.dnd_external_targets and destination_tid:
                source = src_model.get_value(dragged_iter, 0)
                f(source, destination_tid)

    # Separators support ##############################################
    def _separator_func(self, model, itera, user_data=None):
        """ Call user function to determine if this node is separator """
        if itera and model.iter_is_valid(itera):
            node_id = model.get_value(itera, 0)
            node = self.basetree.get_node(node_id)
            if self.separator_func:
                return self.separator_func(node)
            else:
                return False
        else:
            return False

    def set_row_separator_func(self, func, data=None):
        """ Enable support for row separators.

        @param func - function which determines if a node is separator,
            None will disable support for row separators.
        """
        self.separator_func = func
        Gtk.TreeView.set_row_separator_func(self, self._separator_func, data)

    # Multiple selection ####################################################
    def get_selected_nodes(self):
        """ Return list of node_id from liblarch for selected nodes """
        # Get the selection in the Gtk.TreeView
        selection = self.get_selection()
        # Get the selection iter
        if selection.count_selected_rows() <= 0:
            ids = []
        else:
            model, paths = selection.get_selected_rows()
            iters = [model.get_iter(path) for path in paths]
            ts = self.get_model()
            # 0 is the column of the tid
            ids = [ts.get_value(iter, 0) for iter in iters]

        return ids

    def set_multiple_selection(self, multiple_selection):
        """ Allow/forbid multiple selection in TreeView """
        # TODO support for dragging multiple rows at the same time
        # See LP #817433

        if multiple_selection:
            selection_type = Gtk.SelectionMode.MULTIPLE
        else:
            selection_type = Gtk.SelectionMode.SINGLE

        self.get_selection().set_mode(selection_type)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Liblarch - a library to handle directed acyclic graphs
# Copyright (c) 2011-2012 - Lionel Dricot & Izidor Matušov
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------


class _Node(object):
    """ Object just for a single node in Tree """
    def __init__(self, node_id, parent=None):
        """ Initializes node

        @param node_id - unique identifier of node (str)
        @param parent - node_id of parent
        """
        self.node_id = node_id

        self.parents_enabled = True
        self.children_enabled = True
        self.parents = []
        self.children = []

        self.tree = None
        self.pending_relationships = []

        if parent:
            self.add_parent(parent)

    def __str__(self):
        return "<TreeNode: '%s'>" % (self.node_id)

    def get_id(self):
        """ Return node_id """
        return self.node_id

    def modified(self, priority="low"):
        """ Force to update node (because it has changed) """
        if self.tree:
            self.tree.modify_node(self.node_id, priority=priority)

    def _set_tree(self, tree):
        """ Set tree which is should contain this node.

        This method should be called only from MainTree. It is not
        part of public interface. """
        self.tree = tree

    def get_tree(self):
        """ Return associated tree with this node """
        return self.tree

    # Parents #################################################################
    def set_parents_enabled(self, bol):
        if not bol:
            for p in self.get_parents():
                self.remove_parent(p)
        self.parents_enabled = bol

    def has_parents_enabled(self):
        return self.parents_enabled

    def add_parent(self, parent_id):
        """ Add a new parent """
        if (parent_id != self.get_id() and
                self.parents_enabled and
                parent_id not in self.parents):
            if not self.tree:
                self.pending_relationships.append((parent_id, self.get_id()))
            elif not self.tree.has_node(parent_id):
                self.tree.pending_relationships.append(
                    (parent_id, self.get_id()))
            else:
                par = self.tree.get_node(parent_id)
                if par.has_children_enabled():
                    self.tree.new_relationship(parent_id, self.node_id)

    def set_parent(self, parent_id):
        """ Remove other parents and set this parent as only parent """
        if parent_id != self.get_id() and self.parents_enabled:
            is_already_parent_flag = False
            if not self.tree:
                self.pending_relationships.append((parent_id, self.get_id()))
            elif not self.tree.has_node(parent_id):
                for p in self.get_parents():
                    self.tree.break_relationship(p, self.get_id())
                self.tree.pending_relationships.append(
                    (parent_id, self.get_id()))
            else:
                par = self.tree.get_node(parent_id)
                if par.has_children_enabled():
                    # First we remove all the other parents
                    for node_id in self.parents:
                        if node_id != parent_id:
                            self.remove_parent(node_id)
                        else:
                            is_already_parent_flag = True
                    if parent_id and not is_already_parent_flag:
                        self.add_parent(parent_id)

    def remove_parent(self, parent_id):
        """ Remove parent """
        if self.parents_enabled and parent_id in self.parents:
            self.parents.remove(parent_id)
            self.tree.break_relationship(parent_id, self.node_id)

    def has_parent(self, parent_id=None):
        """ Has parent/parents?

        @param parent_id - None => has any parent?
            not None => has this parent?
        """
        if self.parents_enabled:
            if parent_id:
                parent_in_tree = self.tree.has_node(parent_id)
                own_parent = parent_id in self.parents
                return parent_in_tree and own_parent
            else:
                return len(self.parents) > 0
        else:
            return False

    def get_parents(self):
        """ Return parents of node """
        parents = []
        if self.parents_enabled and self.tree:
            for parent_id in self.parents:
                if self.tree.has_node(parent_id):
                    parents.append(parent_id)

        return parents

    # Children ################################################################
    def set_children_enabled(self, bol):
        if not bol:
            for c in self.get_children():
                self.tree.break_relationship(self.get_id(), c)
        self.children_enabled = bol

    def has_children_enabled(self):
        return self.children_enabled

    def add_child(self, child_id):
        """ Add a children to node """
        if self.children_enabled and child_id != self.get_id():
            if child_id not in self.children:
                if not self.tree:
                    self.pending_relationships.append(
                        (self.get_id(), child_id))
                elif not self.tree.has_node(child_id):
                    self.tree.pending_relationships.append(
                        (self.get_id(), child_id))
                else:
                    child = self.tree.get_node(child_id)
                    if child.has_parents_enabled():
                        self.children.append(child_id)
                        self.tree.new_relationship(self.node_id, child_id)
            else:
                print("{} was already in children of {}".format(
                    child_id, self.node_id))

    def has_child(self, child_id=None):
        """ Has child/children?

        @param child_id - None => has any child?
            not None => has this child?
        """
        if self.children_enabled:
            if child_id:
                return child_id in self.children
            else:
                return bool(self.children)
        else:
            return False

    def get_children(self):
        """ Return children of nodes """
        children = []
        if self.children_enabled and self.tree:
            for child_id in self.children:
                if self.tree.has_node(child_id):
                    children.append(child_id)

        return children

    def get_n_children(self):
        """ Return count of children """
        if self.children_enabled:
            return len(self.get_children())
        else:
            return 0

    def get_nth_child(self, index):
        """ Return nth child """
        try:
            return self.children[index]
        except(IndexError):
            raise ValueError("Requested non-existing child")

    def get_child_index(self, node_id):
        if self.children_enabled and node_id in self.children:
            return self.children.index(node_id)
        else:
            return None



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Liblarch - a library to handle directed acyclic graphs
# Copyright (c) 2011-2012 - Lionel Dricot & Izidor Matušov
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import functools

from .filteredtree import FilteredTree


# There should be two classes: for static and for dynamic mode
# There are many conditions, and also we would prevent unallowed modes
class ViewTree(object):
    def __init__(self, maininterface, maintree, filters_bank,
                 name=None, refresh=True, static=False):
        """A ViewTree is the interface that should be used to display Tree(s).

           In static mode, FilteredTree layer is not created.
           (There is no need)

           We connect to MainTree or FilteredTree to get informed about
           changes.  If FilteredTree is used, it is connected to MainTree
           to handle changes and then send id to ViewTree if it applies.

           @param maintree: a Tree object, cointaining all the nodes
           @param filters_bank: a FiltersBank object. Filters can be added
                                dinamically to that.
           @param refresh: if True, this ViewTree is automatically refreshed
                           after applying a filter.
           @param static: if True, this is the view of the complete maintree.
                           Filters cannot be added to such a view.
        """
        self.maininterface = maininterface
        self.__maintree = maintree
        self.__cllbcks = {}
        self.__fbank = filters_bank
        self.static = static

        if self.static:
            self._tree = self.__maintree
            self.__ft = None
            self.__maintree.register_callback(
                'node-added',
                functools.partial(self.__emit, 'node-added'))
            self.__maintree.register_callback(
                'node-deleted',
                functools.partial(self.__emit, 'node-deleted'))
            self.__maintree.register_callback(
                'node-modified',
                functools.partial(self.__emit, 'node-modified'))
        else:
            self.__ft = FilteredTree(
                maintree, filters_bank, name=name, refresh=refresh)
            self._tree = self.__ft
            self.__ft.set_callback(
                'added',
                functools.partial(self.__emit, 'node-added-inview'))
            self.__ft.set_callback(
                'deleted',
                functools.partial(self.__emit, 'node-deleted-inview'))
            self.__ft.set_callback(
                'modified',
                functools.partial(self.__emit, 'node-modified-inview'))
            self.__ft.set_callback(
                'reordered',
                functools.partial(self.__emit, 'node-children-reordered'))

    def queue_action(self, node_id, func, param=None):
        self.__ft.set_callback('runonce', func, node_id=node_id, param=param)

    def get_basetree(self):
        """ Return Tree object """
        return self.maininterface

    def register_cllbck(self, event, func):
        """ Store function and return unique key which can be used to
        unregister the callback later """

        if event not in self.__cllbcks:
            self.__cllbcks[event] = {}

        callbacks = self.__cllbcks[event]
        key = 0
        while key in callbacks:
            key += 1

        callbacks[key] = func
        return key

    def deregister_cllbck(self, event, key):
        """ Remove the callback identifed by key (from register_cllbck) """
        try:
            del self.__cllbcks[event][key]
        except KeyError:
            pass

    def __emit(self, event, node_id, path=None, neworder=None):
        """ Handle a new event from MainTree or FilteredTree
        by passing it to other objects, e.g. TreeWidget """
        callbacks = dict(self.__cllbcks.get(event, {}))
        for func in callbacks.values():
            if neworder:
                func(node_id, path, neworder)
            else:
                func(node_id, path)

    def get_node(self, node_id):
        """ Get a node from MainTree """
        return self.__maintree.get_node(node_id)

    # FIXME Remove this method from public interface
    def get_root(self):
        return self.__maintree.get_root()

    # FIXME Remove this method from public interface
    def refresh_all(self):
        self.__maintree.refresh_all()

    def get_current_state(self):
        """ Request current state to be send by signals/callbacks.

        This allow LibLarch widget to connect on fly (e.g. after FilteredTree
        is up and has some nodes). """

        if self.static:
            self.__maintree.refresh_all()
        else:
            self.__ft.get_current_state()

    def print_tree(self, string=None):
        """ Print the shown tree, i.e. MainTree or FilteredTree """
        return self._tree.print_tree(string)

    def get_all_nodes(self):
        """ Return list of node_id of displayed nodes """
        return self._tree.get_all_nodes()

    def get_n_nodes(self, withfilters=[]):
        """ Returns quantity of displayed nodes in this tree

        @withfilters => Additional filters are applied before counting,
        i.e. the currently applied filters are also taken into account
        """

        if not self.__ft:
            self.__ft = FilteredTree(
                self.__maintree, self.__fbank, refresh=True)
        return self.__ft.get_n_nodes(withfilters=withfilters)

    def get_nodes(self, withfilters=[]):
        """ Returns displayed nodes in this tree

        @withfilters => Additional filters are applied before counting,
        i.e. the currently applied filters are also taken into account
        """

        if not self.__ft:
            self.__ft = FilteredTree(
                self.__maintree, self.__fbank, refresh=True)
        return self.__ft.get_nodes(withfilters=withfilters)

    def get_node_for_path(self, path):
        """ Convert path to node_id.

        I am not sure what this is for... """
        return self._tree.get_node_for_path(path)

    def get_paths_for_node(self, node_id=None):
        """ If node_id is none, return root path

        *Almost* reverse function to get_node_for_path
        (1 node can have many paths, 1:M)
        """
        return self._tree.get_paths_for_node(node_id)

    # FIXME change pid => parent_id
    def next_node(self, node_id, pid=None):
        """ Return the next node to node_id.

        @parent_id => identify which instance of node_id to work.
        If None, random instance is used """

        return self._tree.next_node(node_id, pid)

    def node_has_child(self, node_id):
        """ Has the node at least one child? """
        if self.static:
            return self.__maintree.get_node(node_id).has_child()
        else:
            return self.__ft.node_has_child(node_id)

    def node_all_children(self, node_id=None):
        """ Return children of a node """
        if self.static:
            if not node_id or self.__maintree.has_node(node_id):
                return self.__maintree.get_node(node_id).get_children()
            else:
                return []
        else:
            return self._tree.node_all_children(node_id)

    def node_n_children(self, node_id=None, recursive=False):
        """ Return quantity of children of node_id.
        If node_id is None, use the root node.
        Every instance of node has the same children"""
        if not self.__ft:
            self.__ft = FilteredTree(
                self.__maintree, self.__fbank, refresh=True)
        return self.__ft.node_n_children(node_id, recursive)

    def node_nth_child(self, node_id, n):
        """ Return nth child of the node. """
        if self.static:
            if not node_id or node_id == 'root':
                node = self.__maintree.get_root()
            else:
                node = self.__maintree.get_node(node_id)

            if node and node.get_n_children() > n:
                return node.get_nth_child(n)
            else:
                raise ValueError(
                    "node {} has less than {} nodes".format(node_id, n))
        else:
            realn = self.__ft.node_n_children(node_id)
            if realn <= n:
                raise ValueError(
                    "viewtree has {} nodes, no node {}".format(realn, n))
            return self.__ft.node_nth_child(node_id, n)

    def node_has_parent(self, node_id):
        """ Has node parents? Is it child of root? """
        return len(self.node_parents(node_id)) > 0

    def node_parents(self, node_id):
        """ Returns displayed parents of the given node, or [] if there is no
        parent (such as if the node is a child of the virtual root),
        or if the parent is not displayable.
        Doesn't check wheter node node_id is displayed or not.
        (we only care about parents)
        """
        if self.static:
            return self.__maintree.get_node(node_id).get_parents()
        else:
            return self.__ft.node_parents(node_id)

    def is_displayed(self, node_id):
        """ Is the node displayed? """
        if self.static:
            return self.__maintree.has_node(node_id)
        else:
            return self.__ft.is_displayed(node_id)

    # FILTERS #################################################################
    def list_applied_filters(self):
        return self.__ft.list_applied_filters()

    def apply_filter(self, filter_name, parameters=None,
                     reset=False, refresh=True):
        """ Applies a new filter to the tree.

        @param filter_name: The name of an already registered filter to apply
        @param parameters: Optional parameters to pass to the filter
        @param reset: optional boolean. Should we remove other filters?
        @param refresh : should we refresh after applying this filter ?
        """
        if self.static:
            raise Exception("WARNING: filters cannot be applied"
                            "to a static tree\n")

        self.__ft.apply_filter(filter_name, parameters, reset, refresh)

    def unapply_filter(self, filter_name, refresh=True):
        """ Removes a filter from the tree.

        @param filter_name: The name of filter to remove
        """
        if self.static:
            raise Exception("WARNING: filters cannot be unapplied"
                            "from a static tree\n")

        self.__ft.unapply_filter(filter_name, refresh)

    def reset_filters(self, refresh=True):
        """ Remove all filters currently set on the tree. """
        if self.static:
            raise Exception("WARNING: filters cannot be reset"
                            "on a static tree\n")
        else:
            self.__ft.reset_filters(refresh)
